public class Keccak {

	private static long[] KeccakRoundConstants = new long[24];
	private static int[] KeccakRhoOffsets = new int[25];
	private static byte[] state;
	private static byte[] dataQueue = new byte[144];
	private static int rate, bitsInQueue, squeezing;
	private static byte LFSRstate;

	private static int index(int x, int y) {
		return (((x) % 5) + 5 * ((y) % 5));
	}

	private static long ROL64(long a, int offset) {
		return ((offset != 0) ? ((((long) a) << offset) ^ (((long) a) >>> (64 - offset))) : a);
	}

	public static void Init() {

		KeccakInitialize();
		rate = 1088;
		state = new byte[200];
		bitsInQueue = 0;
		squeezing = 0;

	}

	private static void KeccakInitialize() {
		KeccakInitializeRoundConstants();
		KeccakInitializeRhoOffsets();
	}

	private static void KeccakInitializeRoundConstants() {
		LFSRstate = 0x01;
		int bitPosition;

		for (int i = 0; i < 24; i++) {
			KeccakRoundConstants[i] = 0;
			for (int j = 0; j < 7; j++) {
				bitPosition = (1 << j) - 1;
				if (LFSR86540() != 0)
					KeccakRoundConstants[i] ^= (long) 1 << bitPosition;
			}
		}

	}

	private static int LFSR86540() {
		int result = ((LFSRstate) & 0x01);
		if ((LFSRstate & 0x80) != 0)
			LFSRstate = (byte) ((LFSRstate << 1) ^ 0x71);
		else
			LFSRstate <<= 1;
		return result;
	}

	private static void KeccakInitializeRhoOffsets() {
		int x, y, newX, newY;

		KeccakRhoOffsets[index(0, 0)] = 0;
		x = 1;
		y = 0;
		for (int t = 0; t < 24; t++) {
			KeccakRhoOffsets[index(x, y)] = ((t + 1) * (t + 2) / 2) % 64;
			newX = (0 * x + 1 * y) % 5;
			newY = (2 * x + 3 * y) % 5;
			x = newX;
			y = newY;
		}

	}

	public static void Hash(byte[] data, long databitlen, byte[] hashval) {

		Init();
		Update(data, databitlen);
		Final(hashval);

	}

	public static void Update(byte[] data, long databitlen) {
		long partialBlock, partialByte, wholeBlocks;
		int lastByte;
		int i;

		byte[] curData = null;
		int curDataPointer = 0;

		if ((bitsInQueue % 8) != 0)
			return;
		if (squeezing != 0)
			return;
		i = 0;
		while (i < databitlen) {
			if ((bitsInQueue == 0) && (databitlen >= rate) && (i <= (databitlen - rate))) {

				wholeBlocks = (databitlen - i) / rate;
				curData = new byte[data.length - i / 8];
				System.arraycopy(data, i / 8, curData, 0, curData.length);
				if (rate == 1024) {
					for (int j = 0; j < wholeBlocks; j++, curDataPointer += 1024 / 8) {
						KeccakAbsorb1024bits(state, curData, curDataPointer);
					}
				} else {
					for (int j = 0; j < wholeBlocks; j++, curDataPointer += rate / 8) {
						KeccakAbsorb(state, curData, curDataPointer, rate / 64);
					}
				}
				i += wholeBlocks * rate;
			} else {
				partialBlock = databitlen - i;
				if (partialBlock + bitsInQueue > rate)
					partialBlock = rate - bitsInQueue;
				partialByte = (partialBlock % 8);
				partialBlock -= partialByte;

				System.arraycopy(data, i / 8, dataQueue, bitsInQueue / 8, (int) (partialBlock / 8));

				bitsInQueue += partialBlock;
				i += partialBlock;
				if (bitsInQueue == rate)
					AbsorbQueue();

				if (partialByte > 0) {
					lastByte = ((255 & data[i / 8]) >>> (8 - partialByte));
					dataQueue[bitsInQueue / 8] = (byte) lastByte;
					bitsInQueue += partialByte;
					i += partialByte;
				}
			}
		}
	}

	public static void Final(byte[] hashval) {
		PadAndSwitchToSqueezingPhase();
		System.arraycopy(dataQueue, 0, hashval, 0, 32);

	}

	private static void PadAndSwitchToSqueezingPhase() {
		if ((bitsInQueue % 8) != 0) {
			int padByte = 1 << (bitsInQueue % 8);
			dataQueue[bitsInQueue / 8] |= padByte;
			bitsInQueue += 8 - (bitsInQueue % 8);
		} else {
			dataQueue[bitsInQueue / 8] = 0x01;
			bitsInQueue += 8;
		}
		if (bitsInQueue == rate)
			AbsorbQueue();
		dataQueue[bitsInQueue / 8] = (byte) (32 & 255);
		bitsInQueue += 8;
		if (bitsInQueue == rate)
			AbsorbQueue();
		dataQueue[bitsInQueue / 8] = (byte) ((rate / 8) & 255);
		bitsInQueue += 8;
		if (bitsInQueue == rate)
			AbsorbQueue();
		dataQueue[bitsInQueue / 8] = 0x01;
		bitsInQueue += 8;
		if (bitsInQueue > 0)
			AbsorbQueue();
		KeccakExtract(state, dataQueue, rate / 64);
		squeezing = 1;

	}

	private static void KeccakExtract(final byte[] state, byte[] data, int laneCount) {
		System.arraycopy(state, 0, data, 0, laneCount * 8);
	}

	private static void AbsorbQueue() {
		for (int i = 0; i < (rate / 8) - (bitsInQueue / 8); i++)
			dataQueue[i + (bitsInQueue / 8)] = 0x00;
		KeccakAbsorb(state, dataQueue, 0, rate / 64);
		bitsInQueue = 0;

	}

	private static void KeccakAbsorb(byte[] state, final byte[] data, int offset, int laneCount) {
		KeccakPermutationAfterXor(state, data, offset, laneCount * 8);
	}

	private static void KeccakAbsorb1024bits(byte[] state, final byte[] data, int offset) {
		KeccakPermutationAfterXor(state, data, offset, 128);
	}

	private static void KeccakPermutationAfterXor(byte[] state, final byte[] data, int offset,
			int dataLengthInBytes) {
		for (int i = 0; i < dataLengthInBytes; i++)
			state[i] ^= data[i + offset];
		KeccakPermutation(state);
	}

	private static void KeccakPermutation(byte[] state) {
		long[] stateLong = new long[state.length / 8];
		stateToLong(stateLong);
		KeccakPermutationOnWords(stateLong);
		stateFromLong(stateLong);
	}

	private static void stateToLong(long[] stateLong) {
		int a = 0;
		long value;
		for (int j = 0; j < stateLong.length; j++) {
			value = (long) (255 & state[a++]) << 0;
			value |= (long) (255 & state[a++]) << 8;
			value |= (long) (255 & state[a++]) << 16;
			value |= (long) (255 & state[a++]) << 24;
			value |= (long) (state[a++] & 255) << 32;
			value |= (long) (state[a++] & 255) << 40;
			value |= (long) (state[a++] & 255) << 48;
			value |= (long) (state[a++] & 255) << 56;
			stateLong[j] = value;
		}

	}

	private static void stateFromLong(long[] stateLong) {
		int a = 0;
		for (int i = 0; i < stateLong.length; i++) {
			state[a++] = (byte) (stateLong[i] >>> 0);
			state[a++] = (byte) (stateLong[i] >>> 8);
			state[a++] = (byte) (stateLong[i] >>> 16);
			state[a++] = (byte) (stateLong[i] >>> 24);
			state[a++] = (byte) (stateLong[i] >>> 32);
			state[a++] = (byte) (stateLong[i] >>> 40);
			state[a++] = (byte) (stateLong[i] >>> 48);
			state[a++] = (byte) (stateLong[i] >>> 56);
		}
	}

	private static void KeccakPermutationOnWords(long[] state) {
		for (int i = 0; i < 24; i++) {

			theta(state);
			rho(state);
			pi(state);
			chi(state);
			iota(state, i);
		}
	}

	private static void theta(long[] A) {

		long[] C = new long[5];
		long[] D = new long[5];

		for (int x = 0; x < 5; x++) {
			C[x] = 0;
			for (int y = 0; y < 5; y++)
				C[x] ^= A[index(x, y)];
			D[x] = ROL64(C[x], 1);
		}
		for (int x = 0; x < 5; x++)
			for (int y = 0; y < 5; y++)
				A[index(x, y)] ^= D[(x + 1) % 5] ^ C[(x + 4) % 5];
	}

	private static void rho(long[] A) {

		for (int x = 0; x < 5; x++)
			for (int y = 0; y < 5; y++)
				A[index(x, y)] = ROL64(A[index(x, y)], KeccakRhoOffsets[index(x, y)]);
	}

	private static void pi(long[] A) {
		long[] tempA = new long[25];

		for (int x = 0; x < 5; x++)
			for (int y = 0; y < 5; y++)
				tempA[index(x, y)] = A[index(x, y)];
		for (int x = 0; x < 5; x++)
			for (int y = 0; y < 5; y++)
				A[index(0 * x + 1 * y, 2 * x + 3 * y)] = tempA[index(x, y)];
	}

	private static void chi(long[] A) {
		long[] C = new long[5];

		for (int y = 0; y < 5; y++) {
			for (int x = 0; x < 5; x++)
				C[x] = A[index(x, y)] ^ ((~A[index(x + 1, y)]) & A[index(x + 2, y)]);
			for (int x = 0; x < 5; x++)
				A[index(x, y)] = C[x];
		}
	}

	private static void iota(long[] A, int indexRound) {
		A[index(0, 0)] ^= KeccakRoundConstants[indexRound];
	}

}
